"""
To execute every module directly - remove relative imports "."
"""

from . import machine_state_watcher
from . import privileges
from . import windows
from . import workstation


__version__ = "0.0.2"
